#include <u.h>
#include <libc.h>
#include <regexp.h>
#include <bio.h>
#include <ctype.h>


void
usage(void)
{
	fprint(2,"usage: %s [-x pattern]\n",argv0);
	exits("usage");
}

/*
 * trail length in utf8
 * we must prevent buffer over run
 */
int
traillen(uchar *buf, uchar *end)
{
	int k;
	for(k=0; &buf[k] < end; k++)
		if((buf[k] & 0xc0) != 0x80)
			return k;
	return -1;
}


/*
 * utf8 check: returns position at non-utf8
 *    0bbbbbbb
 *    110bbbbb, 10bbbbbb
 *    1110bbbb, 10bbbbbb, 10bbbbbb
 *    11110bbb, 10bbbbbb, 10bbbbbb, 10bbbbbb
 */
int
chk(uchar *buf, int m)
{
	int k;
	uchar *end = &buf[m];
	for(k = 0; k < m;){
		if(iscntrl(buf[k]) && (buf[k] != '\n') && (buf[k] != '\r') && (buf[k] != '\t'))
			return k;
		if((buf[k] & 0x80) == 0){
			k += 1;
			continue;
		}
		if((buf[k] & 0xe0) == 0xa0)
			if(traillen(&buf[k+1],end) == 1){
				k += 2;
				continue;
			}
		if((buf[k] & 0xf0) == 0xe0)
			if(traillen(&buf[k+1],end) == 2){
				k += 3;
				continue;
			}
		if((buf[k] & 0xf8) == 0xf0)
			if(traillen(&buf[k+1],end) == 3){
				k += 4;
				continue;
			}
		return k;
	}
	return k;
}

void
main(int argc, char *argv[])
{
	int m, fd;
	char *s, buf[512];
	char *xp = nil;
	Biobuf bin;
	Reprog *pat = nil;
	ARGBEGIN{
	case 'x': xp = EARGF(usage());
		break;
	default: usage();
	}ARGEND

	if(argc != 0)
		usage();
	if(xp)
		pat = regcomp(xp);
	Binit(&bin, 0, OREAD);
	while(1){
		s = Brdstr(&bin, '\n', 1);
		if(s == nil)
			break;
		// print("# %s\n", s); /* DBG */
		/* usage example: /sys/src/cmd/exportfs/pattern.c */
		if(pat && regexec(pat, s, nil, 0))
			continue;
		fd = open(s, OREAD);
		if(fd < 0)
			continue;
		m = read(fd, buf, sizeof buf);
		close(fd);
		if(chk((uchar*)buf, m) < m - 4)
			continue;;
		print("%s\n", s);
	}
	exits(nil);
}
